@@ -18,6 +18,7 @@
 
 #include <functional>
 #include <memory>
+#include <optional>
 
 #include <QElapsedTimer>
 #include <QIcon>
@@ -31,7 +32,7 @@
 
 class QMouseEvent;
 
-namespace KDecoration2
+namespace KDecoration3
 {
 class Decoration;
 }
@@ -52,10 +53,17 @@ class WindowItem;
 
 namespace Decoration
 {
-class DecoratedClientImpl;
+class DecoratedWindowImpl;
 class DecorationPalette;
 }
 
+using ElectricBorderMode = std::variant<QuickTileMode, MaximizeMode>;
+
+/**
+ * The PlacementCommand type specifies how a window should be placed in the workspace when it's shown.
+ */
+using PlacementCommand = std::variant<QPointF, QRectF, MaximizeMode>;
+
 class KWIN_EXPORT Window : public QObject
 {
     Q_OBJECT
@@ -386,10 +394,15 @@ class KWIN_EXPORT Window : public QObject
 
     /**
      * The Caption of the Window. Read from WM_NAME property together with a suffix for hostname and shortcut.
-     * To read only the caption as provided by WM_NAME, use the getter with an additional @c false value.
+     * To read only the caption as provided by WM_NAME, use @c captionNormal.
      */
     Q_PROPERTY(QString caption READ caption NOTIFY captionChanged)
 
+    /**
+     * The Caption of the Window. Read from WM_NAME property.
+     */
+    Q_PROPERTY(QString captionNormal READ captionNormal NOTIFY captionNormalChanged)
+
     /**
      * Minimum size as specified in WM_NORMAL_HINTS
      */
@@ -452,7 +465,7 @@ class KWIN_EXPORT Window : public QObject
      * The decision whether a window has a border or not belongs to the window manager.
      * If this property gets abused by application developers, it will be removed again.
      */
-    Q_PROPERTY(bool noBorder READ noBorder WRITE setNoBorder)
+    Q_PROPERTY(bool noBorder READ noBorder WRITE setNoBorder NOTIFY noBorderChanged)
 
     /**
      * Whether the Window provides context help. Mostly needed by decorations to decide whether to
@@ -467,6 +480,14 @@ class KWIN_EXPORT Window : public QObject
      */
     Q_PROPERTY(bool maximizable READ isMaximizable)
 
+    /**
+     * Whether the window is maximized horizontally, vertically or fully.
+     * This is read only, in order to maximize from a script use
+     * the setMaximize function
+     * @see setMaximize
+     */
+    Q_PROPERTY(KWin::MaximizeMode maximizeMode READ maximizeMode NOTIFY maximizedChanged)
+
     /**
      * Whether the Window is moveable. Even if it is not moveable, it might be possible to move
      * it to another screen. The property is evaluated each time it is invoked.
@@ -535,7 +556,7 @@ class KWIN_EXPORT Window : public QObject
     /**
      * The Tile this window is associated to, if any
      */
-    Q_PROPERTY(KWin::Tile *tile READ tile WRITE setTile NOTIFY tileChanged)
+    Q_PROPERTY(KWin::Tile *tile READ requestedTile WRITE setTileCompatibility NOTIFY tileChanged)
 
     /**
      * Returns whether this window is a input method window.
@@ -543,6 +564,18 @@ class KWIN_EXPORT Window : public QObject
      */
     Q_PROPERTY(bool inputMethod READ isInputMethod)
 
+    /**
+     * A client-provided tag of the window.
+     * Not necessarily unique, but can be used to identify similar windows
+     * across application restarts
+     */
+    Q_PROPERTY(QString tag READ tag NOTIFY tagChanged)
+
+    /**
+     * A client-provided description of the window
+     */
+    Q_PROPERTY(QString description READ description NOTIFY descriptionChanged)
+
 public:
     ~Window() override;
 
@@ -628,10 +661,12 @@ public:
      * Calculates the matching client position for the given frame position @p point.
      */
     virtual QPointF framePosToClientPos(const QPointF &point) const;
+    virtual QPointF nextFramePosToClientPos(const QPointF &point) const;
     /**
      * Calculates the matching frame position for the given client position @p point.
      */
     virtual QPointF clientPosToFramePos(const QPointF &point) const;
+    virtual QPointF nextClientPosToFramePos(const QPointF &point) const;
     /**
      * Calculates the matching client size for the given frame size @p size.
      *
@@ -640,6 +675,7 @@ public:
      * Default implementation returns the frame size with frame margins being excluded.
      */
     virtual QSizeF frameSizeToClientSize(const QSizeF &size) const;
+    virtual QSizeF nextFrameSizeToClientSize(const QSizeF &size) const;
     /**
      * Calculates the matching frame size for the given client size @p size.
      *
@@ -648,18 +684,21 @@ public:
      * Default implementation returns the client size with frame margins being included.
      */
     virtual QSizeF clientSizeToFrameSize(const QSizeF &size) const;
+    virtual QSizeF nextClientSizeToFrameSize(const QSizeF &size) const;
     /**
      * Calculates the matching client rect for the given frame rect @p rect.
      *
      * Notice that size constraints won't be applied.
      */
     QRectF frameRectToClientRect(const QRectF &rect) const;
+    QRectF nextFrameRectToClientRect(const QRectF &rect) const;
     /**
      * Calculates the matching frame rect for the given client rect @p rect.
      *
      * Notice that size constraints won't be applied.
      */
     QRectF clientRectToFrameRect(const QRectF &rect) const;
+    QRectF nextClientRectToFrameRect(const QRectF &rect) const;
 
     /**
      * How to resize the window in order to obey constraints (mainly aspect ratios).
@@ -674,17 +713,39 @@ public:
     virtual QSizeF constrainClientSize(const QSizeF &size, SizeMode mode = SizeModeAny) const;
     QSizeF constrainFrameSize(const QSizeF &size, SizeMode mode = SizeModeAny) const;
 
-    void move(const QPointF &point);
+    /**
+     * Moves the window so that the new topLeft corner of the frame is @p topLeft.
+     */
+    void move(const QPointF &topLeft);
+
+    /**
+     * Resizes the window to have a new @p size but stay with the top-left corner in the same position.
+     */
     void resize(const QSizeF &size);
-    void moveResize(const QRectF &rect);
+
+    /**
+     * Requests a new @p geometry for the window that the implementation will need to adopt
+     * within its possibilities.
+     */
+    void moveResize(const QRectF &geometry);
+
+    /**
+     * Returns @c true if the window has already been placed in the workspace; otherwise returns @c false.
+     */
+    bool isPlaced() const;
+
+    /**
+     * Places the window in the workspace as specified by the @a placement command.
+     */
+    void place(const PlacementCommand &placement);
 
     void growHorizontal();
     void shrinkHorizontal();
     void growVertical();
     void shrinkVertical();
 
-    virtual QRectF resizeWithChecks(const QRectF &geometry, const QSizeF &s) = 0;
-    QRectF keepInArea(QRectF geometry, QRectF area, bool partial = false);
+    virtual QRectF resizeWithChecks(const QRectF &geometry, const QSizeF &s) const = 0;
+    QRectF keepInArea(QRectF geometry, QRectF area, bool partial = false) const;
 
     // prefer isXXX() instead
     virtual WindowType windowType() const = 0;
@@ -757,8 +818,8 @@ public:
     bool readyForPainting() const; // true if the window has been already painted its contents
     void setOpacity(qreal opacity);
     qreal opacity() const;
-    virtual bool setupCompositing();
-    virtual void finishCompositing();
+    bool setupCompositing();
+    void finishCompositing();
     EffectWindow *effectWindow();
     const EffectWindow *effectWindow() const;
     SurfaceItem *surfaceItem() const;
@@ -912,7 +973,6 @@ public:
 
     void cancelAutoRaise();
 
-    virtual void updateMouseGrab();
     /**
      * @returns The caption consisting of captionNormal and captionSuffix
      * @see captionNormal
@@ -940,16 +1000,6 @@ public:
     void setHiddenByShowDesktop(bool hidden);
     Window *findModal() const;
     virtual bool isTransient() const;
-    /**
-     * @returns Whether there is a hint available to place the Window on it's parent, default @c false.
-     * @see transientPlacementHint
-     */
-    virtual bool hasTransientPlacementHint() const;
-    /**
-     * Only valid id hasTransientPlacementHint is true
-     * @returns The position the transient wishes to position itself
-     */
-    virtual QRectF transientPlacement() const;
     const Window *transientFor() const;
     Window *transientFor();
     void setTransientFor(Window *transientFor);
@@ -979,7 +1029,7 @@ public:
     }
     void setShortcut(const QString &cut);
 
-    virtual QRectF iconGeometry() const;
+    QRectF iconGeometry() const;
 
     void setMinimized(bool set);
     bool isMinimized() const
@@ -999,17 +1049,18 @@ public:
     virtual void setFullScreen(bool set);
 
     bool wantsAdaptiveSync() const;
+    bool wantsTearing(bool tearingRequested) const;
 
     QRectF geometryRestore() const;
     void setGeometryRestore(const QRectF &rect);
     virtual bool isMaximizable() const;
     virtual MaximizeMode maximizeMode() const;
     virtual MaximizeMode requestedMaximizeMode() const;
-    virtual void maximize(MaximizeMode mode);
+    virtual void maximize(MaximizeMode mode, const QRectF &restore = QRectF());
     /**
      * Sets the maximization according to @p vertically and @p horizontally.
      */
-    Q_INVOKABLE void setMaximize(bool vertically, bool horizontally);
+    Q_INVOKABLE void setMaximize(bool vertically, bool horizontally, const QRectF &restore = QRectF());
 
     QPalette palette();
     const Decoration::DecorationPalette *decorationPalette();
@@ -1059,22 +1110,10 @@ public:
     bool wantsTabFocus() const;
     virtual bool takeFocus() = 0;
     virtual bool wantsInput() const = 0;
-    /**
-     * Whether a dock window wants input.
-     *
-     * By default KWin doesn't pass focus to a dock window unless a force activate
-     * request is provided.
-     *
-     * This method allows to have dock windows take focus also through flags set on
-     * the window.
-     *
-     * The default implementation returns @c false.
-     */
-    virtual bool dockWantsInput() const;
     void checkWorkspacePosition(QRectF oldGeometry = QRectF(), const VirtualDesktop *oldDesktop = nullptr);
     virtual xcb_timestamp_t userTime() const;
 
-    void keyPressEvent(uint key_code);
+    void keyPressEvent(QKeyCombination key_code);
 
     virtual void pointerEnterEvent(const QPointF &globalPos);
     virtual void pointerLeaveEvent();
@@ -1083,18 +1122,20 @@ public:
     void packTo(qreal left, qreal top);
 
     Tile *tile() const;
-    void setTile(Tile *tile);
+    void commitTile(Tile *tile);
+    Tile *requestedTile() const;
+    void requestTile(Tile *tile);
+    void forgetTile(Tile *tile);
+    void setTileCompatibility(Tile *tile);
 
-    /**
-     * Sets the quick tile mode ("snap") of this window.
-     * This will also handle preserving and restoring of window geometry as necessary.
-     * @param mode The tile mode (left/right) to give this window.
-     * @param keyboard Defines whether to take keyboard cursor into account.
-     */
-    void setQuickTileMode(QuickTileMode mode, bool keyboard = false);
+    void handleQuickTileShortcut(QuickTileMode mode);
+    void setQuickTileModeAtCurrentPosition(QuickTileMode mode);
+    void setQuickTileMode(QuickTileMode mode, const QPointF &tileAtPoint);
     QuickTileMode quickTileMode() const;
     QuickTileMode requestedQuickTileMode() const;
 
+    void handleCustomQuickTileShortcut(QuickTileMode mode);
+
     Layer layer() const;
     void updateLayer();
 
@@ -1129,11 +1170,12 @@ public:
     }
     uint32_t interactiveMoveResizeCount() const;
 
-    void updateInteractiveMoveResize(const QPointF &global);
+    void updateInteractiveMoveResize(const QPointF &global, Qt::KeyboardModifiers modifiers);
     /**
      * Ends move resize when all pointer buttons are up again.
      */
     void endInteractiveMoveResize();
+    void cancelInteractiveMoveResize();
 
     virtual StrutRect strutRect(StrutArea area) const;
     StrutRects strutRects() const;
@@ -1144,32 +1186,31 @@ public:
 
     /**
      * Determines the mouse command for the given @p button in the current state.
-     *
-     * The @p handled argument specifies whether the button was handled or not.
-     * This value should be used to determine whether the mouse button should be
-     * passed to the Window or being filtered out.
      */
-    Options::MouseCommand getMouseCommand(Qt::MouseButton button, bool *handled) const;
-    Options::MouseCommand getWheelCommand(Qt::Orientation orientation, bool *handled) const;
-    bool performMouseCommand(Options::MouseCommand, const QPointF &globalPos);
+    std::optional<Options::MouseCommand> getMousePressCommand(Qt::MouseButton button) const;
+    std::optional<Options::MouseCommand> getMouseReleaseCommand(Qt::MouseButton button) const;
+    std::optional<Options::MouseCommand> getWheelCommand(Qt::Orientation orientation) const;
+    bool mousePressCommandConsumesEvent(Options::MouseCommand command) const;
+    /**
+     * @returns whether or not the command consumes the event that triggered it
+     */
+    bool performMousePressCommand(Options::MouseCommand, const QPointF &globalPos);
+    void performMouseReleaseCommand(Options::MouseCommand, const QPointF &globalPos);
 
     // decoration related
     Qt::Edge titlebarPosition() const;
     bool titlebarPositionUnderMouse() const;
-    KDecoration2::Decoration *decoration()
-    {
-        return m_decoration.decoration.get();
-    }
-    const KDecoration2::Decoration *decoration() const
+    KDecoration3::Decoration *decoration() const
     {
         return m_decoration.decoration.get();
     }
+    virtual KDecoration3::Decoration *nextDecoration() const;
     bool isDecorated() const
     {
         return m_decoration.decoration != nullptr;
     }
-    Decoration::DecoratedClientImpl *decoratedClient() const;
-    void setDecoratedClient(Decoration::DecoratedClientImpl *client);
+    Decoration::DecoratedWindowImpl *decoratedWindow() const;
+    void setDecoratedWindow(Decoration::DecoratedWindowImpl *client);
     bool decorationHasAlpha() const;
     void triggerDecorationRepaint();
     void layoutDecorationRects(QRectF &left, QRectF &top, QRectF &right, QRectF &bottom) const;
@@ -1324,8 +1365,9 @@ public:
     void unrefOffscreenRendering();
     bool isOffscreenRendering() const;
 
-    qreal preferredBufferScale() const;
-    void setPreferredBufferScale(qreal scale);
+    qreal targetScale() const;
+    qreal nextTargetScale() const;
+    void setNextTargetScale(qreal scale);
 
     OutputTransform preferredBufferTransform() const;
     void setPreferredBufferTransform(OutputTransform transform);
@@ -1333,6 +1375,9 @@ public:
     const ColorDescription &preferredColorDescription() const;
     void setPreferredColorDescription(const ColorDescription &description);
 
+    QString tag() const;
+    QString description() const;
+
 public Q_SLOTS:
     virtual void closeWindow() = 0;
 
@@ -1402,6 +1447,7 @@ Q_SIGNALS:
      * This signal is emitted when associated tile has changed, including from and to none
      */
     void tileChanged(KWin::Tile *tile);
+    void requestedTileChanged();
 
     void fullScreenChanged();
     void skipTaskbarChanged();
@@ -1449,6 +1495,11 @@ Q_SIGNALS:
     void maximizeGeometryRestoreChanged();
     void fullscreenGeometryRestoreChanged();
     void offscreenRenderingChanged();
+    void targetScaleChanged();
+    void nextTargetScaleChanged();
+    void noBorderChanged();
+    void tagChanged();
+    void descriptionChanged();
 
 protected:
     Window();
@@ -1461,6 +1512,7 @@ protected:
     void autoRaise();
     bool isMostRecentlyRaised() const;
     void markAsDeleted();
+    void markAsPlaced();
     /**
      * Whether the window accepts focus.
      * The difference to wantsInput is that the implementation should not check rules and return
@@ -1529,7 +1581,7 @@ protected:
     virtual void doSetHiddenByShowDesktop();
     virtual void doSetSuspended();
     virtual void doSetModal();
-    virtual void doSetPreferredBufferScale();
+    virtual void doSetNextTargetScale();
     virtual void doSetPreferredBufferTransform();
     virtual void doSetPreferredColorDescription();
 
@@ -1543,8 +1595,8 @@ protected:
     bool isActiveFullScreen() const;
 
     // electric border / quick tiling
-    void setElectricBorderMode(QuickTileMode mode);
-    QuickTileMode electricBorderMode() const
+    void setElectricBorderMode(std::optional<ElectricBorderMode> mode);
+    std::optional<ElectricBorderMode> electricBorderMode() const
     {
         return m_electricMode;
     }
@@ -1556,14 +1608,14 @@ protected:
     void updateElectricGeometryRestore();
     QRectF quickTileGeometryRestore() const;
     QRectF quickTileGeometry(QuickTileMode mode, const QPointF &pos) const;
-    void updateQuickTileMode(QuickTileMode newMode);
+    void exitQuickTileMode();
 
     // geometry handling
     void checkOffscreenPosition(QRectF *geom, const QRectF &screenArea);
-    int borderLeft() const;
-    int borderRight() const;
-    int borderTop() const;
-    int borderBottom() const;
+    qreal borderLeft() const;
+    qreal borderRight() const;
+    qreal borderTop() const;
+    qreal borderBottom() const;
 
     enum class MoveResizeMode : uint {
         None,
@@ -1591,6 +1643,10 @@ protected:
     {
         m_interactiveMoveResize.anchor = anchor;
     }
+    void setInteractiveMoveResizeModifiers(Qt::KeyboardModifiers modifiers)
+    {
+        m_interactiveMoveResize.modifiers = modifiers;
+    }
     /**
      * @returns whether the move resize mode is unrestricted.
      */
@@ -1681,11 +1737,9 @@ protected:
     /**
      * Called during handling a resize. Implementing subclasses can use this
      * method to perform windowing system specific syncing.
-     *
-     * Default implementation does nothing.
      */
     virtual void doInteractiveResizeSync(const QRectF &rect);
-    QRectF titleBarRect(const QRectF &rect, bool &transposed) const;
+    qreal titlebarThickness() const;
     QRectF nextInteractiveMoveGeometry(const QPointF &global) const;
     QRectF nextInteractiveResizeGeometry(const QPointF &global) const;
     void dontInteractiveMoveResize();
@@ -1698,7 +1752,7 @@ protected:
      */
     Gravity mouseGravity() const;
 
-    void setDecoration(std::shared_ptr<KDecoration2::Decoration> decoration);
+    void setDecoration(std::shared_ptr<KDecoration3::Decoration> decoration);
     void startDecorationDoubleClickTimer();
     void invalidateDecorationDoubleClickTimer();
     void updateDecorationInputShape();
@@ -1730,9 +1784,12 @@ protected:
     void cleanTabBox();
     void maybeSendFrameCallback();
 
-    void updatePreferredBufferScale();
+    void updateNextTargetScale();
     void updatePreferredBufferTransform();
     void updatePreferredColorDescription();
+    void setTargetScale(qreal scale);
+
+    void setDescription(const QString &description);
 
     Output *m_output = nullptr;
     QRectF m_frameGeometry;
@@ -1742,7 +1799,8 @@ protected:
     bool m_hidden = false;
     bool m_hiddenByShowDesktop = false;
 
-    qreal m_preferredBufferScale = 1;
+    qreal m_nextTargetScale = 1;
+    qreal m_targetScale = 1;
     OutputTransform m_preferredBufferTransform = OutputTransform::Normal;
     ColorDescription m_preferredColorDescription = ColorDescription::sRGB;
 
@@ -1768,6 +1826,7 @@ protected:
     QIcon m_icon;
     bool m_active = false;
     bool m_deleted = false;
+    bool m_placed = false;
     bool m_keepAbove = false;
     bool m_keepBelow = false;
     bool m_demandsAttention = false;
@@ -1793,14 +1852,13 @@ protected:
     QList<Window *> m_transients;
     bool m_modal = false;
     Layer m_layer = UnknownLayer;
+    QPointer<Tile> m_requestedTile;
     QPointer<Tile> m_tile;
 
     // electric border/quick tiling
-    QuickTileMode m_electricMode = QuickTileFlag::None;
+    std::optional<ElectricBorderMode> m_electricMode = std::nullopt;
     QRectF m_electricGeometryRestore;
     bool m_electricMaximizing = false;
-    // The requested quick tile mode of this window.
-    QuickTileMode m_requestedQuickTileMode = QuickTileFlag::None;
     QTimer *m_electricMaximizingDelay = nullptr;
 
     // geometry
@@ -1816,6 +1874,7 @@ protected:
         bool enabled = false;
         bool unrestricted = false;
         QPointF anchor;
+        Qt::KeyboardModifiers modifiers;
         QPointF offset;
         QRectF initialGeometry;
         QRectF initialGeometryRestore;
@@ -1831,8 +1890,8 @@ protected:
 
     struct
     {
-        std::shared_ptr<KDecoration2::Decoration> decoration;
-        QPointer<Decoration::DecoratedClientImpl> client;
+        std::shared_ptr<KDecoration3::Decoration> decoration;
+        QPointer<Decoration::DecoratedWindowImpl> client;
         QElapsedTimer doubleClickTimer;
         QRegion inputRegion;
     } m_decoration;
@@ -1851,6 +1910,9 @@ protected:
     bool m_lockScreenOverlay = false;
     uint32_t m_offscreenRenderCount = 0;
     QTimer m_offscreenFramecallbackTimer;
+
+    QString m_tag;
+    QString m_description;
 };
 
 inline QRectF Window::bufferGeometry() const
